/* xipcmp - compress a file using the same algorithm as Steve Judd's "xip".
 *
 * Public domain by Russell Marks, 2003.
 */

#include <stdio.h>
#include <stdlib.h>

/* start location of decompressor in speccy's memory */
#define DECMP_ORG	23771

/* and where it decompresses to */
#define DECMP_DEST	0xa000


struct histo_tag
  {
  int idx;
  int count;
  };


int bitbox,bitsleft;


void bit_init(void)
{
bitbox=0; bitsleft=8;
}


void bit_output(int bit)
{
bitsleft--;
bitbox|=(bit<<bitsleft);
if(!bitsleft)
  {
  putchar(bitbox);
  bitbox=0;
  bitsleft=8;
  }
}


void bit_flush()
{
/* there are never 0 bits left outside of bit_output, but
 * if 8 bits are left here there's nothing to flush, so
 * only do it if bitsleft!=8.
 */
if(bitsleft!=8)
  {
  bitsleft=1;
  bit_output(0);	/* yes, really. This will always work. */
  }
}



/* output low n bits of c */
void bit_output_bits(int c,int n)
{
int f;
int mask=(1<<(n-1));

for(f=0;f<n;f++,mask>>=1)
  bit_output((c&mask)?1:0);
}


int histocmp(const void *a1,const void *a2)
{
const struct histo_tag *p1=a1,*p2=a2;

if(p1->count > p2->count)
  return(-1);
if(p1->count < p2->count)
  return(1);
return(0);
}


int main(int argc,char *argv[])
{
static unsigned char buf[65536];
static unsigned char decmpbuf[1024];
static struct histo_tag histo[256];
FILE *in;
int decmpsize,decmp_nval_ofs;
int filesize;
int nval;
int f,g,c;
int bitstotal;
int best_nval,best_size;
int num_lookup;

if(argc!=2)
  {
  fprintf(stderr,"usage: xipcmp xipdecmp.bin <source.bin >compressed.bin\n");
  exit(1);
  }

if((in=fopen(argv[1],"rb"))==NULL)
  {
  fprintf(stderr,"xipcmp: couldn't open %s.\n",argv[1]);
  exit(1);
  }

decmpsize=fread(decmpbuf,1,sizeof(decmpbuf),in);
fclose(in);

decmp_nval_ofs=decmpbuf[decmpsize-1]+1;
decmpsize--;


filesize=fread(buf,1,sizeof(buf),stdin);

for(f=0;f<256;f++)
  {
  histo[f].idx=f;
  histo[f].count=0;
  }

for(f=0;f<filesize;f++)
  histo[buf[f]].count++;

qsort(histo,256,sizeof(struct histo_tag),histocmp);

best_nval=0;
best_size=filesize;

for(nval=2;nval<=8;nval++)
  {
  int size;

  num_lookup=(1<<(nval-1));
  
  /* needed for decmp code and lookup table */
  bitstotal=8*(decmpsize+num_lookup);
  
  for(f=0;f<filesize;f++)
    {
    c=buf[f];
    for(g=0;g<num_lookup;g++)
      if(histo[g].idx==c)
        {
        bitstotal+=nval;
        c=-1;
        break;
        }
    if(c!=-1)
      bitstotal+=9;
    }

  size=(bitstotal+7)/8;

  if(size<best_size)
    best_size=size,best_nval=nval;
  }

if(best_nval==0)
  {
  /* copier code for uncompressed data */
  static unsigned char copycode[11]=
    {
    0x21,(DECMP_ORG+11)%256,(DECMP_ORG+11)/256,	/* ld hl,DECMP_ORG+11 */
    0x11,DECMP_DEST%256,DECMP_DEST/256,		/* ld de,DECMP_DEST */
    0xC1,					/* pop bc */
					/* (this is STACK-BC, i.e. 2d2bh) */
    0xD5,					/* push de */
    0xED,0xB0,					/* ldir */
    0xC9					/* ret */
    };
  
  fprintf(stderr,"xipcmp: warning: couldn't compress, writing with copier\n");
  fwrite(copycode,1,sizeof(copycode),stdout);
  fwrite(buf,1,filesize,stdout);
  exit(0);
  }

fprintf(stderr,"xipcmp: best size %d (n=%d), %d bytes saved\n",
        best_size,best_nval,filesize-best_size);

nval=best_nval;
num_lookup=(1<<(nval-1));

decmpbuf[decmp_nval_ofs]=nval-1;
decmpbuf[1]=(DECMP_ORG+decmpsize+num_lookup)%256;
decmpbuf[2]=(DECMP_ORG+decmpsize+num_lookup)/256;

fwrite(decmpbuf,1,decmpsize,stdout);

for(f=0;f<num_lookup;f++)
  putchar(histo[f].idx);

bit_init();

for(f=0;f<filesize;f++)
  {
  c=buf[f];
  for(g=0;g<num_lookup;g++)
    if(histo[g].idx==c)
      {
      bit_output(0);
      bit_output_bits(g,nval-1);
      c=-1;
      break;
      }
  if(c!=-1)
    {
    bit_output(1);
    bit_output_bits(c,8);
    }
  }

bit_flush();

exit(0);
}
